'use strict'

var Specification = require('./Specification.js')
  , objectConstructor = require('../NativeConstructor.js').objectConstructor
  , ctorUtil = require('../ctorUtil.js')
  , RecordType = require('../types/RecordType.js')
  , RecordProperty = require('../types/RecordProperty.js')
  , evalUtil = require('../evalUtil.js')
  , typeUtil = require('../types/typeUtil.js')
  , DeferredApiCall = require('../DeferredApiCall.js')

function ConstructorSpecification(name) {
  this.isConstructor = true
  Specification.call(this, name)
  this.fn = evalUtil('(function ' + name + '() {})')
  this.base = objectConstructor
  this.baseProps = []
  Object.defineProperty(this, 'recursion', {
    value: false,
    enumerable: true
  })
  this.id = ctorUtil.register(this)
}
ConstructorSpecification.prototype = Object.create(Specification.prototype)
ConstructorSpecification.prototype.constructor = ConstructorSpecification

ConstructorSpecification.prototype.inherit = function(base) {
  if(!(base instanceof ConstructorSpecification))
    throw new Error(
      'Only inheriting from a ConstructorSpecification is supported.')
  this.base = base
  // baseProps are the properties that we let the base constructor func.
  this.baseProps = [].slice.call(arguments, 1)
}

ConstructorSpecification.prototype.toString = function() {
  return 'new ' + Specification.prototype.toString.apply(this, arguments)
}

ConstructorSpecification.prototype.instantiate = function(obj) {
  var result = Object.create(this.fn.prototype)
  var returnType = this.solution ?
    this.solution.type.returnType :
    this.calculateType().returnType
  returnType.children.forEach(function(child) {
    result[child.name] = obj && child.name in obj ?
      obj[child.name] :
      child.type.getDefault()
  })
  if(obj) {
    Object.keys(obj).forEach(function(key) {
      if(key in result) return
      throw new Error('Unexpected property: ' + key)
    })
  }
  return result
}

ConstructorSpecification.prototype.constructsInstance = function(obj) {
  return obj instanceof this.fn ||
    (this.base && this.base.constructsInstance(obj))
}

ConstructorSpecification.prototype.returns = undefined
ConstructorSpecification.prototype.takesContext = undefined

ConstructorSpecification.prototype.returnsRecordThat = function() {

  var self = this
  
  var recordType = DeferredApiCall.callChildren(arguments, function() {
    if(arguments.length)
      throw new Error('Invalid arguments passed to returnsRecordThat: [' +
        [].slice.call(arguments).join(', ') + ']')
    return new RecordType()
  })

  return Specification.prototype.returns.call(this, recordType)
}

ConstructorSpecification.prototype.calculateReturnType = function() {

  var returnType = Specification.prototype.calculateReturnType.apply(
    this, arguments)

  if(!(returnType instanceof RecordType)) {
    throw new Error(
      'Unable to infer a record type for the return value of constructor ' +
      this.name)
  }
  var result = new RecordType()
  result.children = returnType.children.map(function(c) {
    return new RecordProperty(result, c.type, c.name)
  })
  result.ctor = this

  return result
}

ConstructorSpecification.prototype.test =
  function(fn, mocksContainers, suppressRethrow) {
  var ctorFn = !fn ? fn : function() {
    var result = Object.create(fn.prototype)
    fn.apply(result, arguments)
    return result
  }
  return Specification.prototype.test.call(
    this, ctorFn, mocksContainers, suppressRethrow)
}

ConstructorSpecification.prototype.calculateContextType = function() { 
  this.scenarios.forEach(function(s) {
    if(s.calculateContextType())
      throw new Error('Constructor cannot take context')
  })
  return null
}

module.exports = ConstructorSpecification
